# Copyright (C) 2010 W. Trevor King <wking@drexel.edu>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""Various PyRisk class -> graphic renderers.

Creates SVG files by hand.  See the TemplateLibrary class for a
description of map specification format.
"""

import math
import operator
import os
import os.path

from .base import NameMixin, ID_CmpMixin


TEMPLATE_DIRS = [
    './templates',
    '~/share/pyrisk/templates',
    '/usr/share/pyrisk/templates/',
    ]

class Template (NameMixin):
    """Setup regions for a particular world.
    """
    def __init__(self, name, regions, continent_colors={},
                 line_colors={}, player_colors=[]):
        NameMixin.__init__(self, name)
        self.regions = regions
        self.continent_colors = continent_colors
        self.line_colors = line_colors
        self.player_colors = player_colors

class TemplateLibrary (object):
    """Create Templates on demand from a directory of template data.

    TODO: explain template data format.
    """
    def __init__(self, template_dirs=None):
        if template_dirs == None:
            template_dirs = TEMPLATE_DIRS
        self.template_dirs = [os.path.abspath(os.path.expanduser(d))
                              for d in template_dirs]
    def get(self, name):
        region_pointlists,route_pointlists,continent_colors,line_colors, \
            player_colors = \
            self._get_data(name)
        regions = self._generate_regions(region_pointlists, route_pointlists)
        return Template(name, regions, continent_colors, line_colors, 
                        player_colors)
    def _get_data(self, name):
        files = None
        for d in self.template_dirs:
            dirname = os.path.join(d, name.lower())
            try:
                files = os.listdir(dirname)
                break
            except IOError:
                pass
        if files == None:
            return None
        region_pointlists = {}
        route_pointlists = {}
        for filename in files:
            path = os.path.join(dirname, filename)
            name,extension = filename.rsplit('.', 1)
            if extension == 'reg':
                region_pointlists[name] = self._read_pointlist(path)
            elif extension in ['rt', 'vrt']:
                route_pointlists[name] = (self._read_pointlist(path),
                                          extension == 'vrt')
            elif extension == 'col':
                c = self._read_colors(path)
                if name == 'continent':
                    continent_colors = c
                elif name == 'line':
                    line_colors = c
                else:
                    assert name == 'player', name
                    player_colors = []
                    for k,v in sorted(c.items()):
                        player_colors.append(v)
        return (region_pointlists, route_pointlists,
                continent_colors, line_colors, player_colors)
    def _read_pointlist(self, filename):
        pointlist = []
        for line in open(filename, 'r'):
            line = line.strip()
            if len(line) == 0:
                pointlist.append(None)
                continue
            fields = line.split('\t')
            x = int(fields[0])
            y = -int(fields[1])
            if len(fields) == 3:
                label = fields[2].strip()
            else:
                label = None
            pointlist.append((x,y,label))
        return pointlist
    def _read_colors(self, filename):
        colors = {'_attributes': {}}
        for line in open(filename, 'r'):
            line = line.strip()
            if len(line) == 0:
                continue
            fields = line.split('\t')
            name,value = [x.strip() for x in fields]
            if value == '-':
                value = None
            if name.startswith('_'): # attribute setting
                colors['_attributes'][name[1:]] = value
            else: # continent color
                colors[name] = value
        return colors
    def _generate_regions(self, region_pointlists, route_pointlists):
        regions = []
        all_boundaries = []
        for name,pointlist in region_pointlists.items():
            boundaries,head_to_tail = self._pointlist_to_array_of_boundaries(
                all_boundaries, pointlist)
            regions.append(Region(name, boundaries, head_to_tail))
            r = regions[-1]
        for name,v_pointlist in route_pointlists.items():
            pointlist,virtual = v_pointlist
            boundaries,head_to_tail = self._pointlist_to_array_of_boundaries(
                all_boundaries, pointlist)
            assert len(boundaries) == 1, boundaries
            route = boundaries[0]
            route.virtual = virtual
            for terminal in [route[0], route[-1]]:
                for r in regions:
                    for point in r.outline:
                        if hasattr(point, 'name') \
                                and point.name == terminal.name:
                            r.routes.append(route)
                            r.route_head_to_tail.append(
                                terminal == route[0])
                            route.regions.append(r)
        for r in regions:
            r.locate_routes()
        match_counts = [b.match_count for b in all_boundaries]
        assert min(match_counts) in [0, 1], set(match_counts)
        assert max(match_counts) == 1, set(match_counts)
        return regions
    def _pointlist_to_array_of_boundaries(self, all_boundaries, pointlist):
        boundaries = []
        head_to_tail = []
        b_points = []
        for i,point in enumerate(pointlist):
            if point == None:
                boundary,reverse = self._analyze(b_points)
                boundary = self._insert_boundary(all_boundaries, boundary)
                boundaries.append(boundary)
                head_to_tail.append(not reverse)
                b_points = []
                continue
            b_points.append(point)
        if len(b_points) > 0:
            boundary,reverse = self._analyze(b_points)
            boundary = self._insert_boundary(all_boundaries, boundary)
            boundaries.append(boundary)
            head_to_tail.append(not reverse)
        return boundaries, head_to_tail
    def _analyze(self, boundary_points):
        start = self._vbp(boundary_points[0])
        stop = self._vbp(boundary_points[-1])
        if stop < start:
            reverse = True
        points = [self._vbp(b) for b in boundary_points]
        reverse = start > stop
        if reverse == True:
            points.reverse()
            start,stop = (stop, start)
        boundary = Boundary([p-start for p in points])
        for bp,p in zip(boundary, points):
            bp.name = p.name # preserve point names
        boundary.name = '(%s) -> (%s)' % (start.name, stop.name)
        boundary.real_pos = start
        return (boundary, reverse)
    def _vbp(self, boundary_point):
        v = Vector((boundary_point[0], boundary_point[1]))
        v.name = boundary_point[2]
        return v
    def _insert_boundary(self, all_boundaries, new):
        if new in all_boundaries:
            return new
        for b in all_boundaries:
            if len(b) == len(new) and b.real_pos == new.real_pos:
                match = True
                for bp,np in zip(b, new):
                    if bp != np:
                        match = False
                        break
                if match == True:
                    b.match_count += 1
                    return b
        all_boundaries.append(new)
        new.match_count = 0
        return new

TEMPLATE_LIBRARY = TemplateLibrary()


class Vector (tuple):
    """Simple cartesian vector operations.

    >>> v = Vector
    >>> a = v((0, 0))
    >>> b = v((1, 1))
    >>> c = v((2, 3))
    >>> a+b
    (1, 1)
    >>> a+b+c
    (3, 4)
    >>> b-c
    (-1, -2)
    >>> -c
    (-2, -3)
    >>> b*0.5
    (0.5, 0.5)
    >>> c*(1, -1)
    (2, -3)
    >>> a < b
    True
    >>> c > b
    True
    """
    def _set_name(self, new, other=None):
        if hasattr(self, 'name'):
            if self.name == None:
                if hasattr(other, 'name'):
                    new.name = other.name
                    return
            new.name = self.name
        elif hasattr(other, 'name'):
            new.name = other.name
    def __neg__(self):
        new = self.__class__(map(operator.neg, self))
        self._set_name(new)
        return new
    def mag(self):
        """Return the magnitude.
        """
        return math.sqrt(sum([x**2 for x in self]))
    def angle(self):
        """Return the direction (must be in 2D).
        """
        if len(self) != 2:
            raise ValueError('length != 2, %s' % (self))
        return math.atan2(self[1], self[0])
    def __add__(self, other):
        if len(self) != len(other):
            raise ValueError('length missmatch %s, %s' % (self, other))
        new = self.__class__(map(operator.add, self, other))
        self._set_name(new, other)
        return new
    def __sub__(self, other):
        if len(self) != len(other):
            raise ValueError('length missmatch %s, %s' % (self, other))
        new = self.__class__(map(operator.sub, self, other))
        self._set_name(new, other)
        return new
    def __mul__(self, other):
        """Return the elementwise product (with vectors) or scalar
        product (with numbers).
        """
        if hasattr(other, '__iter__'): # dot product
            if len(self) != len(other):
                raise ValueError('length missmatch %s, %s' % (self, other))
            new = self.__class__([s*o for s,o in zip(self, other)])
            #map(operator.mul, self, other))
        else: # scalar product
            new = self.__class__([x*other for x in self])
        self._set_name(new, other)
        return new
    def triple_angle(self, previous, next):
        """Return the angle between (previous->self) and (self->next)
        in radians.
        """
        angle = (next-self).angle() - (self-previous).angle()
        while angle > math.pi:
            angle -= 2*math.pi
        while angle < -math.pi:
            angle += 2*math.pi
        return angle
    def cross(self, other):
        """Return the cross product.
        
        In 2D, just return the z component of the result.
        """
        if len(self) != len(other):
            raise ValueError('length missmatch %s, %s' % (self, other))
        if len(self) not in [2,3]:
            raise ValueError('cross product not defined in %dD, %s, %s'
                             % (len(self), self, other))
        if len(self) == 2:
            s = self.__class__((self[0], self[1], 0))
            o = self.__class__((other[0], other[1], 0))
        else:
            s = self; o = other
        new = self.__class__((s[1]*o[2]-s[2]*o[1],
                              -s[0]*o[2]-s[2]*o[0],
                              s[0]*o[1]-s[1]*o[0]))
        if len(self) == 2:
            return new[2]
        return new

def nameless(vector):
    """Return a nameless version of a given Vector.

    Useful for ensuring the result of a sum / etc. has the name of the
    *other* vector.
    """
    return Vector(vector)

class Boundary (ID_CmpMixin, list):
    """Contains a list of points along the boundary.

    All positions are relative to the location of the first point,
    which should therefore always be (0,0).
    """
    def __init__(self, points):
        list.__init__(self)
        ID_CmpMixin.__init__(self)
        for p in points:
            self.append(Vector(p))
        self.regions = []
        assert self[0] == (0,0), self
        self.x_min = min([p[0] for p in self])
        self.x_max = max([p[0] for p in self])
        self.y_min = min([p[1] for p in self])
        self.y_max = max([p[1] for p in self])

class Route (ID_CmpMixin):
    """Connect non-adjacent Regions.
    """
    def __init__(self, boundary):
        ID_CmpMixin.__init__(self)
        self.boundary = boundary

class Region (NameMixin, ID_CmpMixin, list):
    """Contains a list of boundaries and a label.

    Regions can be Territories or pieces of Territories (e.g separate
    islands, such as in Indonesia).

    Boundaries must be entered counter clockwise (otherwise you
    will get an error).

    >>> r = Region('Earth',
    ...            [Boundary([(0,0), (10,0)]),
    ...             Boundary([(0,0), (0,10)]),
    ...             Boundary([(0,0), (10,0)]),
    ...             Boundary([(0,0), (0,10)])],
    ...            [True, True, False, False])
    >>> r.outline
    [(0, 0), (10, 0), (10, 10), (0, 10), (0, 0)]
    >>> r.area()
    100.0
    >>> r.center_of_mass()
    (5, 5)
    """
    def __init__(self, name, boundaries, head_to_tail, routes=None,
                 route_head_to_tail=None):
        NameMixin.__init__(self, name)
        list.__init__(self, boundaries)
        ID_CmpMixin.__init__(self)
        for boundary in self:
            boundary.regions.append(self)
        self.head_to_tail = head_to_tail
        self.routes = routes
        self.route_head_to_tail = route_head_to_tail
        if routes == None:
            assert route_head_to_tail == None
            self.routes = []
            self.route_head_to_tail = []
        self.route_starts = [] # set by .locate_routes
        self.generate_outline() # sets .outline, .starts
        self.x_min = min([b.x_min+s[0] for b,s in zip(self, self.starts)])
        self.x_max = max([b.x_max+s[0] for b,s in zip(self, self.starts)])
        self.y_min = min([b.y_min+s[1] for b,s in zip(self, self.starts)])
        self.y_max = max([b.y_max+s[1] for b,s in zip(self, self.starts)])
    def generate_outline(self):
        """Return a list of boundary points surrounding the region.

        The main issue here is determining the proper border
        orientation for a CCW outline, which we do via a user-supplied
        list head_to_tail.
        """
        self.starts = []
        points = [Vector((0,0))]
        for boundary,htt in zip(self, self.head_to_tail):
            pos = points[-1]
            if htt == True:
                assert boundary[0] == (0,0), boundary
                new = boundary[1:]
            else:
                pos -= nameless(boundary[-1])
                new = reversed(boundary[:-1])
            pos = nameless(pos)
            self.starts.append(pos)
            for p in new:
                points.append(pos+p)
        assert points[-1] == points[0], '%s: %s (not closed)' % (self, points)
        self.outline = points
        total_angle = 0
        for prev,p,next in self._unique_triples():
            total_angle += p.triple_angle(prev, next)
        assert abs(total_angle-2*math.pi) < 0.1, \
            "%s: %s (not CCW: %.2f)" % (self, points, total_angle)
    def locate_routes(self):
        self.route_starts = []
        for route,htt in zip(self.routes, self.route_head_to_tail):
            if htt:
                anchor = route[0]
            else:
                anchor = route[-1]
            for point in self.outline:
                if hasattr(point, 'name') and point.name == anchor.name:
                    self.route_starts.append(point-nameless(anchor))
                    break
    def area(self):
        """Return the region's enclosed area as a float.
        """
        return sum([self._triangle_area(a,b,c) for a,b,c in self._triangles()])
    def center_of_mass(self):
        """Return a vector locating the region's center of mass.

        Truncated to the lower-left pixel.
        """
        cas = []
        for a,b,c in self._triangles():
            cas.append((self._triangle_center_of_mass(a,b,c),
                        self._triangle_area(a,b,c)))
        m = sum([a for c,a in cas])
        average = Vector((int(sum([c[0]*a for c,a in cas])/m),
                          int(sum([c[1]*a for c,a in cas])/m)))
        return average
    def _unique_triples(self):
        """Iterate through consecutive triples (prev, p, next)
        for every unique point in self.outline.
        """
        unique = self.outline[:-1]
        assert len(unique) >= 3, \
            '%s: %s (< 3 unique points)' % (self, self.outline)
        for i,p in enumerate(unique):
            prev = unique[i-1]
            next = unique[(i+1)%len(unique)]
            yield (prev, p, next)
    def _triangles(self):
        """Iterate through CCW triangles composing the region.

        >>> r = Region('Earth',
        ...            [Boundary([(0,0),(1,-1),(0,3),(-1,0),(0,0)])],
        ...            [True])
        >>> for triangle in r._triangles():
        ...     print triangle
        ((0, 0), (1, -1), (0, 3))
        ((-1, 0), (0, 0), (0, 3))
        """
        points = self.outline[:-1]
        while len(points) >= 3:
            for i,p in enumerate(points):
                prev = points[i-1]
                next = points[(i+1)%len(points)]
                if p.triple_angle(prev, next) > 0:
                    break # found first CCW triangle
            points.remove(p) # drop the outer point
            yield (prev, p, next)
    def _triangle_area(self, a, b, c):
        """
        >>> r = Region('Earth',
        ...            [Boundary([(0,0), (1,0), (1,1), (0,0)])],
        ...            [True])
        >>> triangle = r.outline[:3]
        >>> triangle
        [(0, 0), (1, 0), (1, 1)]
        >>> r._triangle_area(*triangle)
        0.5
        """
        return abs(0.5 * (b-a).cross(c-a))
    def _triangle_center_of_mass(self, a, b, c):
        """
        >>> r = Region('Earth',
        ...            [Boundary([(0,0), (1,0.3), (0.5,1.2), (0,0)])],
        ...            [True])
        >>> triangle = r.outline[:3]
        >>> ['(%.2f, %.2f)' % x for x in triangle]
        ['(0.00, 0.00)', '(1.00, 0.30)', '(0.50, 1.20)']
        >>> r._triangle_center_of_mass(*triangle)
        (0.5, 0.5)
        """
        return (a+b+c)*(1/3.)

class WorldRenderer (object):
    def __init__(self, template_lib=None, line_width=2, buf=10, dpcm=50):
        self.template_lib = template_lib
        if self.template_lib == None:
            self.template_lib = TEMPLATE_LIBRARY
        self.buf = buf
        self.line_width = line_width
        self.dpcm = dpcm
        self.army_scale = 3
    def filename_and_mime_image_type(self, world):
        """Return suggestions for emailing the rendered object.

        Returns (filename, subtype), where the MIME type is
        image/<subtype>.
        """
        return ('%s.svg' % world.name, 'svg+xml') 
    def render(self, world, players):
        template = self.template_lib.get(world.name)
        if template == None:
            template = self._auto_template(world)
        return self.render_template(world, players, template)
    def render_template(self, world, players, template):
        region_pos,width,height = self._locate(template)
        lines = [
            '<?xml version="1.0" standalone="no"?>',
            '<!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 1.1//EN"',
            '  "http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd">',
            '<svg width="%.1fcm" height="%.1fcm" viewBox="0 0 %d %d"'
            % (float(width)/self.dpcm, float(height)/self.dpcm,
               width, height),
            '     xmlns="http://www.w3.org/2000/svg" version="1.1">',
            '<title>%s</title>' % template,
            '<desc>A PyRisk world snapshot</desc>',
           ]
        terr_regions = {}
        drawn_rts = {}
        for r in template.regions:
            t = self._matching_territory(world, r)
            if t.name in terr_regions:
                terr_regions[t.name].append(r)
            else:
                terr_regions[t.name] = [r]
            c_col = template.continent_colors[t.continent.name]
            if len(template.continent_colors['_attributes']) == 0:
                c_col_attr = ''
            else:
                attrs = template.continent_colors['_attributes'].items()
                c_col_attr = ''.join([' %s="%s"' % (k,v) for k,v in attrs])
            if template.line_colors['border'] == None:
                b_col_attr = ''
            else:
                b_col_attr = ' stroke="%s" stroke-width="%d"' \
                    % (template.line_colors['border'], self.line_width)
            lines.extend([
                    '<polygon title="%s / %s / %s"'
                    % (t, t.player, t.armies),                    
                    '         fill="%s"%s%s' % (c_col, c_col_attr, b_col_attr),
                    '         points="%s" />'
                    % ' '.join(['%d,%d' % ((region_pos[id(r)]+p)
                                           *(1,-1) # svg y value increases down
                                           +(0,height)) # shift back into bbox
                                for p in r.outline[:-1]])
                    ])
            for rt,rt_start in zip(r.routes, r.route_starts):
                if id(rt) in drawn_rts:
                    continue
                drawn_rts[id(rt)] = rt
                if rt.virtual == True:
                    color = template.line_colors['virtual route']
                else:
                    color = template.line_colors['route']
                if color == None:
                    continue
                lines.extend([
                        '<polyline stroke="%s" stroke-width="%d"'
                        % (color, self.line_width),
                        '         points="%s" />'
                        % ' '.join(['%d,%d' % ((region_pos[id(r)]+rt_start+p)
                                           *(1,-1) # svg y value increases down
                                           +(0,height)) # shift back into bbox
                                    for p in rt])
                        ])
        for t in world.territories():
            regions = terr_regions[t.name]
            center = self._territory_center(region_pos, regions)
            radius = self.army_scale*math.sqrt(t.armies)
            color = template.player_colors[players.index(t.player)]
            if color == None:
                continue
            if template.line_colors['army'] == None:
                b_col_attr = ''
            else:
                b_col_attr = ' stroke="%s" stroke-width="%d"' \
                    % (template.line_colors['army'], self.line_width)
                radius += self.line_width/2.
            lines.extend([
                    '<circle title="%s / %s / %s"' % (t, t.player, t.armies),
                    '         fill="%s"%s' % (color, b_col_attr),
                    '         cx="%d" cy="%d" r="%.1f" />'
                    % (center[0], center[1]*-1+height, radius),
                    ])
        lines.extend(['</svg>', ''])
        return '\n'.join(lines)
    def _locate(self, template):
        region_pos = {} # {id: absolute position, ...}
        boundary_pos = {} # {id: absolute position, ...}
        route_pos = {} # {id: absolute position, ...}
        b1 = template.regions[0][0]
        boundary_pos[id(b1)] = Vector((0,0)) # fix the first boundary point
        stack = [r for r in b1.regions]
        while len(stack) > 0:
            r = stack.pop()
            if id(r) in region_pos:
                continue # skip duplicate entries
            r_start = None
            for b,rel_b_start in zip(r, r.starts):
                if id(b) in boundary_pos:
                    b_start = boundary_pos[id(b)]
                    r_start = b_start - rel_b_start
                    break # found an anchor
            if r_start == None:
                for rt,rel_rt_start in zip(r.routes, r.route_starts):
                    if id(rt) in route_pos:
                        rt_start = route_pos[id(rt)]
                        r_start = rt_start - rel_rt_start
                        break # found an anchor
            region_pos[id(r)] = r_start
            for b,rel_b_start in zip(r, r.starts):
                if id(b) not in boundary_pos:
                    boundary_pos[id(b)] = r_start + rel_b_start
                    for r2 in b.regions:
                        stack.append(r2)
            for rt,rt_start in zip(r.routes, r.route_starts):
                if id(rt) not in route_pos:
                    route_pos[id(rt)] = r_start + rt_start
                    for r2 in rt.regions:
                        stack.append(r2)
        for r in template.regions:
            if id(r) not in region_pos:
                raise KeyError(r.name)
        x_min = min([r.x_min + region_pos[id(r)][0]
                     for r in template.regions]) - self.buf
        x_max = max([r.x_max + region_pos[id(r)][0]
                     for r in template.regions]) + self.buf
        y_min = min([r.y_min + region_pos[id(r)][1]
                     for r in template.regions]) - self.buf
        y_max = max([r.y_max + region_pos[id(r)][1]
                     for r in template.regions]) + self.buf
        for key,value in region_pos.items():
            region_pos[key] = value - Vector((x_min, y_min))
        return (region_pos, x_max-x_min, y_max-y_min)
    def _matching_territory(self, world, region):
        t = None
        try:
            t = world.territory_by_name(region.name)
        except KeyError:
            for rt in region.routes:
                if not rt.virtual:
                    continue
                for r in rt.regions:
                    try:
                        t = world.territory_by_name(r.name)
                    except KeyError:
                        pass
        assert t != None, 'No territory in %s associated with region %s' \
            % (world, region)
        return t
    def _territory_center(self, region_pos, regions):
        """Return the center of mass of a territory composed of regions.
        """
        cas = []
        for r in regions:
            center = r.center_of_mass() + region_pos[id(r)]
            area = r.area()
            cas.append((center, area))
        m = sum([a for c,a in cas])
        average = Vector((int(sum([c[0]*a for c,a in cas])/m),
                          int(sum([c[1]*a for c,a in cas])/m)))
        return average
    def _auto_template(self, world):
        raise NotImplementedError

def test():
    import doctest, sys
    failures,tests = doctest.testmod(sys.modules[__name__])
    return failures

def render_earth():
    from .base import generate_earth,Player,Engine
    players = [Player('Alice'), Player('Bob'), Player('Charlie'),
               Player('Eve'), Player('Mallory'), Player('Zoe')]
    world = generate_earth()
    e = Engine(world, players)
    e.setup()
    r = WorldRenderer()
    print r.render(e.world, players)
